//
//  Device.h
//  BouncingBallWithOde
//

#ifndef __OpenNiMultiViewer__Device__
#define __OpenNiMultiViewer__Device__

#include <boost/thread/future.hpp>
#include <boost/function.hpp>
#include <boost/noncopyable.hpp>
#include <boost/thread.hpp>
#include <boost/array.hpp>
#include <pcl/point_types.h>
#include <XnCppWrapper.h>
#include <vector>
#include <string>

class Context;
namespace pcl
{
        class PolygonMesh;
}

class Device : private boost::noncopyable
{
public:
        enum Mode
        {
                Default,
                NormalExtraction,
        };
        
        struct TriangularMesh
        {
                std::vector<float>              vertices_;
                std::vector<boost::uint32_t>    indices_;
        };
        
        explicit Device(boost::shared_ptr<Context> context);
        void create_image_generator(const xn::NodeInfo& info, const XnMapOutputMode& map_mode);
        void create_depth_generator(const xn::NodeInfo& info, const XnMapOutputMode& map_mode);
        
        boost::shared_ptr<TriangularMesh> get_triangular_mesh()
        {
                boost::unique_future<boost::shared_ptr<TriangularMesh> > f(boost::move(future_));
                return boost::move(f.get());
        }
        
        bool is_ready_for_normals() const
        {
                return future_.valid() && future_.is_ready();
        }
        
        const xn::ImageGenerator& get_image() const
        {
                return image_;
        }
        
        const xn::DepthGenerator& get_depth() const
        {
                return depth_;
        }
        
        void preprocess(int width, int height);
        int get_cloud_size() const
        {
                return cloud_.size() / (sizeof(XnUInt8) * 4 + sizeof(XnFloat) * 3);
        }
        
        const std::vector<XnUInt8>& get_cloud() const
        {
                return cloud_;
        }
        void create_cloud();
        
        void set_image_info_string(const std::string& info)
        {
                image_info_string_ = info;
        }
        
        void set_depth_info_string(const std::string& info)
        {
                depth_info_string_ = info;
        }

        const std::string& get_image_info_string() const
        {
                return image_info_string_;
        }
        
        const std::string& get_depth_info_string() const
        {
                return depth_info_string_;
        }
        
        void set_mode(Mode mode)
        {
                mode_ = mode;
        }
        
        const Mode get_mode() const
        {
                return mode_;
        }
        
private:
        struct NormalExtractionInputGenerator
        {
                NormalExtractionInputGenerator(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud)
                        : cloud_(cloud) {}
                        
                void operator()(XnUInt32 k, XnFloat* depth) const
                {
                        pcl::PointXYZ& p = cloud_->points[k];
                        p.x = depth[0];
                        p.y = depth[1];
                        p.z = depth[2];
                }
                
                pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_;
        };
        
        struct NormalExtractionNothingGenerator
        {
                void operator()(XnUInt32 k, XnFloat* depth) const
                {
                        // do nothing
                }
        };

        boost::shared_ptr<Context> context_;
        xn::ImageGenerator image_;
        xn::DepthGenerator depth_;
        std::vector<XnUInt8> cloud_;
        int width_;
        int height_;
        int cloud_size_;
        std::vector<float> x_factors_;
        std::vector<float> y_factors_;
        std::string image_info_string_;
        std::string depth_info_string_;
        pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_xyz_;
        NormalExtractionInputGenerator normal_extraction_input_generator_;
        NormalExtractionNothingGenerator normal_extraction_nothing_generator_;
        Mode mode_;
        boost::unique_future<boost::shared_ptr<TriangularMesh> > future_;
        
        void create_factor_tables(int width, int height);
        
        void convert_to_real_world(
                const XnUInt32&         i, 
                const XnUInt32&         j, 
                const XnDepthPixel&     depth, 
                float&                  x, 
                float&                  y, 
                float&                  z
        ) const
        {
                z = -static_cast<float>(depth);
                x = z * x_factors_[i];
                y = z * y_factors_[j];
        }
        
        void create_cloud(
                const xn::DepthMetaData&        depth_meta, 
                const xn::ImageMetaData&        image_meta,
                boost::function<void(XnUInt32 k, XnFloat* depth)> fun
        );
        
        void convert_to_polyton_mesh(
                const pcl::PointCloud<pcl::PointXYZ>::Ptr&      cloud,
                pcl::PolygonMesh&                               triangles
        );
        
        void estimate_normals(
                const pcl::PointCloud<pcl::PointXYZ>::Ptr&      cloud,
                pcl::PointCloud<pcl::Normal>::Ptr&              normals
        );

        boost::shared_ptr<TriangularMesh> calculate_triangular_mesh();
        void extract_ode_data(
                const pcl::PolygonMesh& triangles,
                std::vector<float>&     vertices,
                std::vector<boost::uint32_t>& indices
        );
};

#endif /* defined(__OpenNiMultiViewer__Device__) */
