//
//  PhysicalSimulator.cpp
//  BouncingBallWithOde
//

#include "PhysicalSimulator.h"
#include <boost/move/move.hpp>
#include <boost/make_shared.hpp>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/features/normal_3d.h>
#include <boost/random.hpp>
#include <ctime>

PhysicalSimulator::PhysicalSimulator()
        : contacts_(MaxNumOfContacts)
        , collides_(false)
        , triangular_mesh_(0)
{
        dInitODE();
}

PhysicalSimulator::~PhysicalSimulator()
{
        dCloseODE();
}

void PhysicalSimulator::collide(void(*callback)(void* data, dGeomID o1, dGeomID o2))
{
        space_.collide(0, callback);
}

void PhysicalSimulator::near_callback(void* data, dGeomID o1, dGeomID o2)
{
        int n = dCollide(o1, o2, MaxNumOfContacts, &contacts_[0].geom, sizeof(dContact));
        if ( n > 0 ) {
                collides_ = true;
        } else {
                collides_ = false;
        }
        for ( int i = 0; i < n; ++i ) {
                dContact& contact = contacts_[i];
                contact.surface.mode = dContactBounce | dContactSoftERP;
                contact.surface.mu = dInfinity;
                contact.surface.mu2 = 0;
                contact.surface.bounce = 1;
                contact.surface.bounce_vel = 0.1;
//                contact.surface.soft_cfm = 0.01;
                contact.surface.soft_erp = 1.0;
                dContactJoint joint;
                joint.create(world_, joint_group_, &contact);
                joint.attach(dGeomGetBody(contact.geom.g1), dGeomGetBody(contact.geom.g2));
        }
}

void PhysicalSimulator::clear_balls()
{
        for ( std::size_t i = 0; i < balls_.size(); ++i ) {
                boost::shared_ptr<Ball>& ball = balls_[i];
                ball->geom_.destroy();
        }
        balls_.clear();
}

void PhysicalSimulator::create_world()
{
        const dReal gravity[] = {0, -1, 0};
        world_.setGravity(gravity);
}

void PhysicalSimulator::create_balls()
{
        dMass m;
        m.setZero();
        const dReal mass = 100;
        dReal radius = 30;
        m.setSphereTotal(mass, radius);
        const dReal initial_position[] = {200, 100, -1900};

        const int NumOfBalls = 10;
        balls_.clear();
        dReal velocity = 5;
        boost::mt19937 gen(static_cast<int>(time(0)));
        int max_number = 100;
        boost::uniform_int<> dst(1, max_number);
        boost::variate_generator<boost::mt19937, boost::uniform_int<> > rand(gen, dst);
        for ( std::size_t i = 0; i < NumOfBalls; ++i ) {
                boost::shared_ptr<Ball> ball = boost::make_shared<Ball>();
                
                ball->body_.create(world_);
                ball->body_.setMass(&m);
                ball->body_.setPosition(
                        initial_position[0] + radius * 2 * i - radius * (NumOfBalls - 1),
                        initial_position[1],
                        initial_position[2]
                );
                
                dReal v = velocity;
                if ( i % 2 == 0 ) {
                        v = -velocity;
                }
                int r = rand();
                if ( r <  max_number / 3) {
                        ball->body_.setLinearVel(v, 0, 0);
                } else if ( r < 2 * max_number / 3 ) {
                        ball->body_.setLinearVel(0, v, 0);
                } else {
                        ball->body_.setLinearVel(0, 0, v);
                }
                // The geometry(ball.geom) is associated with the rigid body(ball.body).
                ball->geom_.create(space_, radius);
                ball->geom_.setBody(ball->body_);
                balls_.push_back(boost::move(ball));
        }
}

void PhysicalSimulator::register_triangular_mesh(const boost::shared_ptr<Device::TriangularMesh>& mesh)
{
        dTriMeshDataID data = dGeomTriMeshDataCreate();
        dGeomTriMeshDataBuildSingle(
                data,
                &(mesh->vertices_[0]), 3 * sizeof(float), mesh->vertices_.size() / 3, &(mesh->indices_[0]),
                static_cast<int>(mesh->indices_.size()),
                3 * sizeof(dTriIndex)
        );
        if ( triangular_mesh_)  {
                dGeomDestroy(triangular_mesh_);
        }
        triangular_mesh_ = dCreateTriMesh(space_, data, 0, 0, 0);
}