//
//  Viewer.cpp
//  BouncingBallWithOde
//

#include "Viewer.h"
#include "MatrixStack.h"
#include "GeometryDescriptor.h"
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <stdexcept>
#include <string>
#include <sstream>
#include <ctime>
#include <GLUT/GLUT.h>

int Viewer::viewer_number_ = 0;

namespace
{
        const GLfloat light_ambient[]  = { 0.0f, 0.0f, 0.0f, 1.0f };
        const GLfloat light_diffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
        const GLfloat light_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
        const GLfloat light_position[] = { 0.0f, 0.0f, 10.0f, 0.0f };
        
        const GLfloat mat_ambient[]    = { 0.7f, 0.7f, 0.7f, 1.0f };
        const GLfloat mat_diffuse[]    = { 0.8f, 0.8f, 0.8f, 1.0f };
        const GLfloat mat_specular[]   = { 1.0f, 1.0f, 1.0f, 1.0f };
        const GLfloat high_shininess[] = { 100.0f };
        
        const float IntervalTime = 10;
        
        const boost::chrono::seconds TimerDuration(1);
}

Viewer::Viewer()
        : index_(0)
        , device_()
        , simulator_(new PhysicalSimulator)
        , frame_count_(0)
        , is_extracted_(false)
        , simulator_mutex_()
        , stops_frame_rate_(false)
        , io_service_()
        , timer_(io_service_)
        , displays_polygon_(false)
        , displays_balls_(false)
{
        simulator_->create_world();
        timer_.expires_from_now(TimerDuration);
        timer_.async_wait(boost::bind(&Viewer::set_frame_rate, this));
        ++viewer_number_;
}

void Viewer::near_callback(void* data, dGeomID o1, dGeomID o2)
{
        simulator_->near_callback(data, o1, o2);
}

namespace
{
        inline void create_transform_matrix(const double pos[3], const double rot[12], double matrix[16])
        {
                matrix[0] = rot[0];
                matrix[1] = rot[4];
                matrix[2] = rot[8];
                matrix[3] = 0;
                matrix[4] = rot[1];
                matrix[5] = rot[5];
                matrix[6] = rot[9];
                matrix[7] = 0;
                matrix[8] = rot[2];
                matrix[9] = rot[6];
                matrix[10] = rot[10];
                matrix[11] = 0;
                matrix[12] = pos[0];
                matrix[13] = pos[1];
                matrix[14] = pos[2];
                matrix[15] = 1;
        }
        
	const XnUInt8 Colours[][3] =
	{
		{ 255, 0, 0},
		{ 0, 255, 0},
		{ 0, 0, 255},
		{ 255, 255, 0},
		{ 255, 0, 255},
		{ 0, 255, 255}
	};

        inline void paint_color(int key)
        {
                glColor3d(Colours[key][0], Colours[key][1], Colours[key][2]);
        }
        double matrix[16];

}

void Viewer::display_balls(void (*near_callback)(void* data, dGeomID o1, dGeomID o2))
{
        boost::lock_guard<boost::mutex> lock(simulator_mutex_);
        simulator_->set_collides(false);
        simulator_->empty_join_group();
        simulator_->collide(near_callback);
        simulator_->step(0.5);
        int n = sizeof(Colours) / 3;
        int i = 0;
        for ( auto& ball : simulator_->get_balls() ) {
                i = i % n;
                paint_color(i);
                const dReal* pos = ball->body_.getPosition();
                const dReal* rot = ball->body_.getRotation();
                create_transform_matrix(pos, rot, matrix);
                MatrixStack stack;
                glMultMatrixd(matrix);
                glutSolidSphere(ball->geom_.getRadius(), 16, 16);
                ++i;
        }
}

void Viewer::register_mesh()
{
        mesh_ = device_->get_triangular_mesh();
        {
                boost::lock_guard<boost::mutex> lock(simulator_mutex_);
                simulator_->register_triangular_mesh(mesh_);
        }
        std::cout << "[" << window_id_ << "] Mesh has been calculated.\n";
}

int Viewer::initialize(const XnMapOutputMode& map_mode, int index, const boost::shared_ptr<Device>& device)
{
        width_ = map_mode.nXRes;
        height_ = map_mode.nYRes;
        index_ = index;
        device_ = device;
        device_->preprocess(width_, height_);
        glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
        glutInitWindowPosition(100 + index_ * width_, 100);
	window_id_ = glutCreateWindow((std::string("OpenNI Viewer ") + boost::lexical_cast<std::string>(index_)).c_str());
        glutSetCursor(GLUT_CURSOR_LEFT_ARROW);

        set_viewpoint();
        set_environment();
        io_service_thread_ = boost::thread(boost::bind(&boost::asio::io_service::run, &io_service_));
        return window_id_;
}


void Viewer::set_display_callback(void(*display)())
{
        glutDisplayFunc(display);
}

void Viewer::display_point_cloud(const XnUInt8* buf, int size) const
{
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        MatrixStack stack;
        glInterleavedArrays(GL_C4UB_V3F, 0, buf);
        glDrawArrays(GL_POINTS, 0, size);
}

namespace
{
        void draw_square()
        {
                glColor3f(0, 0, 0);
                GeometryDescriptor des(GL_QUADS);
                glVertex3f(0, -10, -1);
                glVertex3f(150, -10, -1);
                glVertex3f(150, 30, -1);
                glVertex3f(0, 30, -1);
        }

        void draw_stroke_string(GLfloat x, GLfloat y, GLfloat z, float xscl, float yscl,void *font, const char *string)
        {
                MatrixStack stack;
                glTranslatef(x, y, z);
                draw_square();
                glScalef(xscl, yscl, 1.0f);
                glColor3d(1, 1, 1);
                while ( *string ) {
                        glutStrokeCharacter(font, *string++);
                }
        }
        
}

void Viewer::set_frame_rate()
{
        frame_rate_ = frame_count_;
        frame_count_ = 0;
        if ( !stops_frame_rate_ ) {
                timer_.expires_from_now(TimerDuration);
                timer_.async_wait(boost::bind(&Viewer::set_frame_rate, this));
        }
}

void Viewer::display_frame_rate()
{
        ostringstream oss;
        oss << frame_rate_ << "(fps)";
        draw_stroke_string(400, -410, -1000, 0.2, 0.2, GLUT_STROKE_MONO_ROMAN, oss.str().c_str());
        //std::cout << "[" << window_id_ << "] " << frame_rate_ << "(fps)" << std::endl;
}

namespace
{

        void draw_mesh(const boost::shared_ptr<Device::TriangularMesh>& mesh)
        {
                const int stride = 3;
                for (int i = 0; i < mesh->indices_.size() / 3; i++) {
                        const float *p0 = &mesh->vertices_[stride * mesh->indices_[i * 3 + 0]];
                        const float* p1 = &mesh->vertices_[stride * mesh->indices_[i * 3 + 1]];
                        const float* p2 = &mesh->vertices_[stride * mesh->indices_[i * 3 + 2]];
                        glColor3f(1, 1, 0);
                        glLineWidth(2);
                        glDisable(GL_LIGHTING);
                        GeometryDescriptor geom(GL_TRIANGLES);
                        glVertex3f(p0[0], p0[1], p0[2]);
                        glVertex3f(p1[0], p1[1], p1[2]);
                        glVertex3f(p2[0], p2[1], p2[2]);
                }
        }
        
        
}

void Viewer::display(void (*near_callback)(void* data, dGeomID o1, dGeomID o2))
{
        device_->create_cloud();
        if ( device_->is_ready_for_normals() ) {
                register_mesh();
                is_extracted_ = true;
        }
        
        display_point_cloud(&device_->get_cloud()[0], device_->get_cloud_size());
        if ( is_extracted_ && displays_polygon_ ) {
                draw_mesh(mesh_);
        }
        
        if ( displays_balls_ && is_extracted_ ) {
                display_balls(near_callback);
        }
        ++frame_count_;
        display_frame_rate();
        glutSwapBuffers();
}

void Viewer::set_viewpoint() const
{
	glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        float ar = static_cast<float>(width_) / height_;
        gluPerspective(46, ar, 1, 5000);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
}

void Viewer::quit()
{
        std::cout << "[" << window_id_ << "] quit\n";
        glutDestroyWindow(window_id_);
        stops_frame_rate_ = true;
        io_service_thread_.join();
        --viewer_number_;
        if ( viewer_number_ == 0 ) {
                std::exit(1);
        }
}

void Viewer::create_mesh()
{
        {
                boost::lock_guard<boost::mutex> lock(simulator_mutex_);
                simulator_->clear_balls();
        }
        is_extracted_ = false;
        displays_balls_ = false;
        std::cout << "[" << window_id_ << "] Mesh is now being calculated.\n";
        device_->set_mode(Device::NormalExtraction);
}

void Viewer::create_balls()
{
        if ( is_extracted_ ) {
                std::cout << "[" << window_id_ << "] Balls are created.\n";
                displays_balls_ = true;
                register_balls();
        }
}

void Viewer::register_balls()
{
        boost::lock_guard<boost::mutex> lock(simulator_mutex_);
        simulator_->create_balls();
}

void Viewer::display_polygon()
{
        displays_polygon_ = !displays_polygon_;
}

void Viewer::on_key(unsigned char key, int /*x*/, int /*y*/)
{
	switch ( key )
	{
	case 'q':
                quit();
                break;
                
	case 'm':
                create_mesh();
		break;

	case 'b':
                create_balls();
                break;
                
        case 'p':
                display_polygon();
                break;
                        
        default:
                break;
	}
}

void Viewer::set_environment() const
{
        glEnable(GL_DEPTH_TEST);
        glClearColor(1, 1, 1, 1);
        glEnable(GL_CULL_FACE);
        glCullFace(GL_BACK);
        glDepthFunc(GL_LESS);

        glEnable(GL_LIGHT0);
        glEnable(GL_NORMALIZE);
        glEnable(GL_COLOR_MATERIAL);
        
        glLightfv(GL_LIGHT0, GL_AMBIENT,  light_ambient);
        glLightfv(GL_LIGHT0, GL_DIFFUSE,  light_diffuse);
        glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
        glLightfv(GL_LIGHT0, GL_POSITION, light_position);

        glMaterialfv(GL_FRONT, GL_AMBIENT,   mat_ambient);
        glMaterialfv(GL_FRONT, GL_DIFFUSE,   mat_diffuse);
        glMaterialfv(GL_FRONT, GL_SPECULAR,  mat_specular);
        glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);
}